﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="DataBindingOperator.cs" company="Slash Games">
//   Copyright (c) Slash Games. All rights reserved.
// </copyright>
// --------------------------------------------------------------------------------------------------------------------

namespace Slash.Unity.DataBind.Core.Presentation
{
    using System.Collections.Generic;
    using System.Diagnostics.CodeAnalysis;
    using System.Linq;
    using UnityEngine;

    /// <summary>
    ///     Base class of a behaviour that uses on one or more data bindings.
    /// </summary>
    public abstract class DataBindingOperator : MonoBehaviour, IContextOperator
    {
        /// <summary>
        ///     Bindings this operator depends on.
        /// </summary>
        private readonly List<DataBinding> bindings = new List<DataBinding>();

        /// <summary>
        ///     Indicates if Init method was called.
        /// </summary>
        private bool isInitialized;

        /// <summary>
        ///     Indicates if the provider listens to value changes of its bindings.
        /// </summary>
        private bool isMonitoringBindings;

        /// <summary>
        ///     Indicates if Start() method was called by Unity.
        ///     Used to call first Enable() not before Awake() of other scripts was called.
        /// </summary>
        private bool isStarted;

        /// <summary>
        ///     Indicates if the data provider already holds a valid value.
        /// </summary>
        public bool IsInitialized
        {
            get
            {
                return this.isInitialized && this.isStarted && (this.bindings == null ||
                                                                this.bindings.All(binding => binding.IsInitialized));
            }
        }

        /// <summary>
        ///     Deinitializes the operator.
        /// </summary>
        public virtual void Deinit()
        {
        }

        /// <summary>
        ///     Disable the operator.
        /// </summary>
        public virtual void Disable()
        {
            this.UnregisterFromValueChanges();
        }

        /// <summary>
        ///     Enable the operator.
        /// </summary>
        public virtual void Enable()
        {
            this.RegisterForValueChanges();
            var bindingsInitialized = this.bindings.All(binding => binding.IsInitialized);
            if (bindingsInitialized)
            {
                this.OnBindingValuesChanged();
            }
        }

        /// <summary>
        ///     Initializes the operator.
        /// </summary>
        public virtual void Init()
        {
        }

        /// <summary>
        ///     Has to be called when an anchestor context changed as the data value may change.
        /// </summary>
        public virtual void OnContextChanged()
        {
            foreach (var binding in this.bindings)
            {
                binding.OnContextChanged();
            }
        }

        /// <summary>
        ///     Adds and initializes the specified binding.
        /// </summary>
        /// <param name="binding">Binding to add.</param>
        protected void AddBinding(DataBinding binding)
        {
            // Init.
            binding.Init(this);

            if (this.isMonitoringBindings)
            {
                binding.ValueChanged += this.OnBindingValueChanged;
            }

            this.bindings.Add(binding);
        }

        /// <summary>
        ///     Adds and initializes the specified bindings.
        /// </summary>
        /// <param name="newBindings">Bindings to add.</param>
        protected void AddBindings(IEnumerable<DataBinding> newBindings)
        {
            foreach (var binding in newBindings)
            {
                this.AddBinding(binding);
            }
        }

        /// <summary>
        ///     Unity callback.
        ///     Overwrite Init method in derived classes to add bindings and do other initialization stuff.
        /// </summary>
        [SuppressMessage("ReSharper", "UnusedMemberHiearchy.Global")]
        protected void Awake()
        {
            this.Init();
            this.isInitialized = true;
        }

        /// <summary>
        ///     Called when a value of the bindings of this operator changed.
        /// </summary>
        protected virtual void OnBindingValuesChanged()
        {
        }

        /// <summary>
        ///     Unity callback.
        ///     Overwrite Deinit method in derived classes to remove bindings and do other deinitialization stuff.
        /// </summary>
        [SuppressMessage("ReSharper", "UnusedMemberHiearchy.Global")]
        protected void OnDestroy()
        {
            this.Deinit();
        }

        /// <summary>
        ///     Unity callback.
        ///     Overwrite Disable method in derived classes to disable behaviour.
        /// </summary>
        [SuppressMessage("ReSharper", "UnusedMemberHiearchy.Global")]
        protected void OnDisable()
        {
            this.Disable();
        }

        /// <summary>
        ///     Unity callback.
        ///     Overwrite Enable method in derived classes to enable behaviour.
        /// </summary>
        [SuppressMessage("ReSharper", "UnusedMemberHiearchy.Global")]
        protected void OnEnable()
        {
            // Don't call Enable when the Start() method was not called,
            // otherwise it may be called before the Awake() method of other scripts was executed.
            // See https://forum.unity.com/threads/onenable-before-awake.361429/ for details.
            if (this.isStarted)
            {
                this.Enable();
            }
        }

        /// <summary>
        ///     Removes and deinitializes the specified binding.
        /// </summary>
        /// <param name="binding">Binding to remove.</param>
        protected void RemoveBinding(DataBinding binding)
        {
            if (this.isMonitoringBindings)
            {
                binding.ValueChanged -= this.OnBindingValueChanged;
            }

            // Deinit.
            binding.Deinit();

            this.bindings.Remove(binding);
        }

        /// <summary>
        ///     Unity callback.
        /// </summary>
        [SuppressMessage("ReSharper", "UnusedMemberHiearchy.Global")]
        protected virtual void Start()
        {
            this.isStarted = true;

            // Enable for the first time.
            this.Enable();
        }

        private void OnBindingValueChanged(object newValue)
        {
            this.OnBindingValuesChanged();
        }

        private void RegisterForValueChanges()
        {
            if (this.isMonitoringBindings)
            {
                return;
            }

            foreach (var binding in this.bindings)
            {
                binding.ValueChanged += this.OnBindingValueChanged;
            }
            this.isMonitoringBindings = true;
        }

        private void UnregisterFromValueChanges()
        {
            if (!this.isMonitoringBindings)
            {
                return;
            }

            foreach (var binding in this.bindings)
            {
                binding.ValueChanged -= this.OnBindingValueChanged;
            }
            this.isMonitoringBindings = false;
        }
    }
}