﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="CollectionRangeLookup.cs" company="Slash Games">
//   Copyright (c) Slash Games. All rights reserved.
// </copyright>
// --------------------------------------------------------------------------------------------------------------------

namespace Slash.Unity.DataBind.Foundation.Providers.Lookups
{
    using System.Linq;
    using Slash.Unity.DataBind.Core.Data;
    using Slash.Unity.DataBind.Core.Presentation;
    using UnityEngine;

    /// <summary>
    ///     Returns a part of a given collection.
    /// </summary>
    [AddComponentMenu("Data Bind/Foundation/Lookups/[DB] Collection Range Lookup")]
    public class CollectionRangeLookup : DataProvider
    {
        /// <summary>
        ///     Collection to get the items from.
        /// </summary>
        [Tooltip("Collection to get item from.")]
        public DataBinding Collection;

        private Collection dataCollection;

        /// <summary>
        ///     Index of the first item to get from the collection.
        /// </summary>
        [Tooltip("Index of the first item to get from the collection.")]
        public DataBinding FirstIndex;

        /// <summary>
        ///     Index of the last item to get from the collection.
        /// </summary>
        [Tooltip("Index of the last item to get from the collection.")]
        public DataBinding LastIndex;

        /// <inheritdoc />
        public override object Value
        {
            get
            {
                if (this.dataCollection == null)
                {
                    return null;
                }

                // Select value range.
                var collection = new Collection<object>();

                var first = this.FirstIndex.GetValue<int>();
                var last = this.LastIndex.GetValue<int>();

                foreach (var item in this.dataCollection.Cast<object>().Skip(first).Take(last - first + 1))
                {
                    collection.Add(item);
                }

                return collection;
            }
        }

        private Collection DataCollection
        {
            get
            {
                return this.dataCollection;
            }
            set
            {
                if (value == this.dataCollection)
                {
                    return;
                }

                if (this.dataCollection != null)
                {
                    this.dataCollection.ItemAdded -= this.OnCollectionItemAdded;
                    this.dataCollection.ItemRemoved -= this.OnCollectionItemRemoved;
                    this.dataCollection.Cleared -= this.OnCollectionCleared;
                }

                this.dataCollection = value;

                if (this.dataCollection != null)
                {
                    this.dataCollection.ItemAdded += this.OnCollectionItemAdded;
                    this.dataCollection.ItemRemoved += this.OnCollectionItemRemoved;
                    this.dataCollection.Cleared += this.OnCollectionCleared;
                }

                this.UpdateValue();
            }
        }

        /// <inheritdoc />
        public override void Deinit()
        {
            this.RemoveBinding(this.FirstIndex);
            this.RemoveBinding(this.LastIndex);
            this.RemoveBinding(this.Collection);
        }

        /// <inheritdoc />
        public override void Disable()
        {
            base.Disable();
            this.Collection.ValueChanged -= this.OnDataCollectionChanged;
        }

        /// <inheritdoc />
        public override void Enable()
        {
            base.Enable();

            this.Collection.ValueChanged += this.OnDataCollectionChanged;
            this.OnDataCollectionChanged(null);
        }

        /// <inheritdoc />
        public override void Init()
        {
            this.AddBinding(this.FirstIndex);
            this.AddBinding(this.LastIndex);
            this.AddBinding(this.Collection);
        }

        /// <inheritdoc />
        protected override void UpdateValue()
        {
            this.OnValueChanged();
        }

        private void OnCollectionCleared()
        {
            this.UpdateValue();
        }

        private void OnCollectionItemAdded(object item)
        {
            this.UpdateValue();
        }

        private void OnCollectionItemRemoved(object item)
        {
            this.UpdateValue();
        }

        private void OnDataCollectionChanged(object newValue)
        {
            this.DataCollection = this.Collection.GetValue<Collection>();
        }
    }
}